"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
Object.defineProperty(exports, "__esModule", { value: true });
var transform_1 = require("../utils/transform");
var style_utils_1 = require("../utils/style-utils");
var deep_equal_1 = require("../utils/deep-equal");
var pointerEvents = {
    mousedown: 'onMouseDown',
    mouseup: 'onMouseUp',
    mouseover: 'onMouseOver',
    mousemove: 'onMouseMove',
    click: 'onClick',
    dblclick: 'onDblClick',
    mouseenter: 'onMouseEnter',
    mouseleave: 'onMouseLeave',
    mouseout: 'onMouseOut',
    contextmenu: 'onContextMenu',
    touchstart: 'onTouchStart',
    touchend: 'onTouchEnd',
    touchmove: 'onTouchMove',
    touchcancel: 'onTouchCancel'
};
var cameraEvents = {
    movestart: 'onMoveStart',
    move: 'onMove',
    moveend: 'onMoveEnd',
    dragstart: 'onDragStart',
    drag: 'onDrag',
    dragend: 'onDragEnd',
    zoomstart: 'onZoomStart',
    zoom: 'onZoom',
    zoomend: 'onZoomEnd',
    rotatestart: 'onRotateStart',
    rotate: 'onRotate',
    rotateend: 'onRotateEnd',
    pitchstart: 'onPitchStart',
    pitch: 'onPitch',
    pitchend: 'onPitchEnd',
    boxzoomstart: 'onBoxZoomStart',
    boxzoomend: 'onBoxZoomEnd',
    boxzoomcancel: 'onBoxZoomCancel'
};
var otherEvents = {
    wheel: 'onWheel',
    resize: 'onResize',
    load: 'onLoad',
    render: 'onRender',
    idle: 'onIdle',
    remove: 'onRemove',
    data: 'onData',
    styledata: 'onStyleData',
    sourcedata: 'onSourceData'
};
var settingNames = [
    'minZoom',
    'maxZoom',
    'minPitch',
    'maxPitch',
    'maxBounds',
    'projection',
    'renderWorldCopies'
];
var handlerNames = [
    'scrollZoom',
    'boxZoom',
    'dragRotate',
    'dragPan',
    'keyboard',
    'doubleClickZoom',
    'touchZoomRotate',
    'touchPitch'
];
/**
 * A wrapper for mapbox-gl's Map class
 */
var Mapbox = /** @class */ (function () {
    function Mapbox(MapClass, props) {
        var _this = this;
        // mapboxgl.Map instance. Not using type here because we are accessing
        // private members and methods
        this._map = null;
        // Internal states
        this._internalUpdate = false;
        this._inRender = false;
        this._hoveredFeatures = null;
        this._moved = false;
        this._zoomed = false;
        this._pitched = false;
        this._rotated = false;
        this._onEvent = function (e) {
            // @ts-ignore
            var cb = _this.props[otherEvents[e.type]];
            if (cb) {
                cb(e);
            }
        };
        this._onError = function (e) {
            if (_this.props.onError) {
                _this.props.onError(e);
            }
            else {
                // eslint-disable-next-line
                console.error(e.error);
            }
        };
        this._onPointerEvent = function (e) {
            // @ts-ignore
            var cb = _this.props[pointerEvents[e.type]];
            if (cb) {
                if (_this.props.interactiveLayerIds && e.type !== 'mouseover' && e.type !== 'mouseout') {
                    var features = _this._hoveredFeatures ||
                        _this._map.queryRenderedFeatures(e.point, {
                            layers: _this.props.interactiveLayerIds
                        });
                    if (!features.length) {
                        return;
                    }
                    e.features = features;
                }
                cb(e);
                delete e.features;
            }
        };
        this._onCameraEvent = function (e) {
            if (_this._internalUpdate) {
                return;
            }
            // @ts-ignore
            var cb = _this.props[cameraEvents[e.type]];
            if (cb) {
                cb(e);
            }
        };
        this._MapClass = MapClass;
        this.props = props;
    }
    Object.defineProperty(Mapbox.prototype, "map", {
        get: function () {
            return this._map;
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(Mapbox.prototype, "viewState", {
        get: function () {
            return (0, transform_1.transformToViewState)(this._renderTransform);
        },
        enumerable: false,
        configurable: true
    });
    Mapbox.prototype.setProps = function (props) {
        if (this._inRender) {
            this._nextProps = props;
            return;
        }
        var oldProps = this.props;
        this.props = props;
        var settingsChanged = this._updateSettings(props, oldProps);
        if (settingsChanged) {
            this._renderTransform = this._map.transform.clone();
        }
        var viewStateChanged = this._updateViewState(props, true);
        this._updateStyle(props, oldProps);
        this._updateHandlers(props, oldProps);
        // If 1) view state has changed to match props and
        //    2) the props change is not triggered by map events,
        // it's driven by an external state change. Redraw immediately
        if (settingsChanged || (viewStateChanged && !this._map.isMoving())) {
            this.redraw();
        }
    };
    Mapbox.prototype.initialize = function (container) {
        var _this = this;
        var props = this.props;
        var mapOptions = __assign(__assign(__assign({}, props), props.initialViewState), { accessToken: props.mapboxAccessToken || getAccessTokenFromEnv() || null, container: container, style: (0, style_utils_1.normalizeStyle)(props.mapStyle) });
        var viewState = mapOptions.initialViewState || mapOptions.viewState || mapOptions;
        Object.assign(mapOptions, {
            center: [viewState.longitude || 0, viewState.latitude || 0],
            zoom: viewState.zoom || 0,
            pitch: viewState.pitch || 0,
            bearing: viewState.bearing || 0
        });
        if (props.gl) {
            // eslint-disable-next-line
            var getContext_1 = HTMLCanvasElement.prototype.getContext;
            // Hijack canvas.getContext to return our own WebGLContext
            // This will be called inside the mapboxgl.Map constructor
            // @ts-expect-error
            HTMLCanvasElement.prototype.getContext = function () {
                // Unhijack immediately
                HTMLCanvasElement.prototype.getContext = getContext_1;
                return props.gl;
            };
        }
        var map = new this._MapClass(mapOptions);
        if (viewState.padding) {
            map.setPadding(viewState.padding);
        }
        this._renderTransform = map.transform.clone();
        // Hack
        // Insert code into map's render cycle
        var renderMap = map._render;
        map._render = this._render.bind(this, renderMap);
        var runRenderTaskQueue = map._renderTaskQueue.run;
        map._renderTaskQueue.run = function (arg) {
            runRenderTaskQueue.call(map._renderTaskQueue, arg);
            _this._onBeforeRepaint();
        };
        // Insert code into map's event pipeline
        var fireEvent = map.fire;
        map.fire = this._fireEvent.bind(this, fireEvent);
        // add listeners
        for (var eventName in pointerEvents) {
            map.on(eventName, this._onPointerEvent);
        }
        for (var eventName in cameraEvents) {
            map.on(eventName, this._onCameraEvent);
        }
        for (var eventName in otherEvents) {
            map.on(eventName, this._onEvent);
        }
        map.on('error', this._onError);
        this._map = map;
    };
    Mapbox.prototype.destroy = function () {
        this._map.remove();
    };
    // Force redraw the map now. Typically resize() and jumpTo() is reflected in the next
    // render cycle, which is managed by Mapbox's animation loop.
    // This removes the synchronization issue caused by requestAnimationFrame.
    Mapbox.prototype.redraw = function () {
        var map = this._map;
        // map._render will throw error if style does not exist
        // https://github.com/mapbox/mapbox-gl-js/blob/fb9fc316da14e99ff4368f3e4faa3888fb43c513
        //   /src/ui/map.js#L1834
        if (map.style) {
            // cancel the scheduled update
            if (map._frame) {
                map._frame.cancel();
                map._frame = null;
            }
            // the order is important - render() may schedule another update
            map._render();
        }
    };
    // Adapted from map.jumpTo
    /* Update camera to match props
       @param {object} nextProps
       @param {bool} triggerEvents - should fire camera events
       @returns {bool} true if anything is changed
     */
    Mapbox.prototype._updateViewState = function (nextProps, triggerEvents) {
        if (this._internalUpdate) {
            return false;
        }
        var map = this._map;
        var tr = this._renderTransform;
        // Take a snapshot of the transform before mutation
        var zoom = tr.zoom, pitch = tr.pitch, bearing = tr.bearing;
        var changed = (0, transform_1.applyViewStateToTransform)(tr, __assign(__assign({}, (0, transform_1.transformToViewState)(map.transform)), nextProps));
        if (changed && triggerEvents) {
            // Delay DOM control updates to the next render cycle
            this._moved = true;
            this._zoomed = this._zoomed || zoom !== tr.zoom;
            this._rotated = this._rotated || bearing !== tr.bearing;
            this._pitched = this._pitched || pitch !== tr.pitch;
        }
        // Avoid manipulating the real transform when interaction/animation is ongoing
        // as it would interfere with Mapbox's handlers
        if (!map.isMoving()) {
            (0, transform_1.applyViewStateToTransform)(map.transform, nextProps);
        }
        return changed;
    };
    /* Update camera constraints and projection settings to match props
       @param {object} nextProps
       @param {object} currProps
       @returns {bool} true if anything is changed
     */
    Mapbox.prototype._updateSettings = function (nextProps, currProps) {
        var _a;
        var map = this._map;
        var changed = false;
        for (var _i = 0, settingNames_1 = settingNames; _i < settingNames_1.length; _i++) {
            var propName = settingNames_1[_i];
            if (!(0, deep_equal_1.deepEqual)(nextProps[propName], currProps[propName])) {
                changed = true;
                (_a = map["set".concat(propName[0].toUpperCase()).concat(propName.slice(1))]) === null || _a === void 0 ? void 0 : _a.call(map, nextProps[propName]);
            }
        }
        return changed;
    };
    /* Update map style to match props
       @param {object} nextProps
       @param {object} currProps
       @returns {bool} true if style is changed
     */
    Mapbox.prototype._updateStyle = function (nextProps, currProps) {
        if (nextProps.cursor !== currProps.cursor) {
            this._map.getCanvas().style.cursor = nextProps.cursor;
        }
        if (nextProps.mapStyle !== currProps.mapStyle) {
            var options = {
                diff: nextProps.styleDiffing
            };
            if ('localIdeographFontFamily' in nextProps) {
                options.localIdeographFontFamily = nextProps.localIdeographFontFamily;
            }
            this._map.setStyle((0, style_utils_1.normalizeStyle)(nextProps.mapStyle), options);
            return true;
        }
        return false;
    };
    /* Update interaction handlers to match props
       @param {object} nextProps
       @param {object} currProps
       @returns {bool} true if anything is changed
     */
    Mapbox.prototype._updateHandlers = function (nextProps, currProps) {
        var map = this._map;
        var changed = false;
        for (var _i = 0, handlerNames_1 = handlerNames; _i < handlerNames_1.length; _i++) {
            var propName = handlerNames_1[_i];
            var newValue = nextProps[propName];
            if (!(0, deep_equal_1.deepEqual)(newValue, currProps[propName])) {
                changed = true;
                if (newValue) {
                    map[propName].enable(newValue);
                }
                else {
                    map[propName].disable();
                }
            }
        }
        return changed;
    };
    Mapbox.prototype._updateHover = function (e) {
        var _a;
        var props = this.props;
        var shouldTrackHoveredFeatures = props.interactiveLayerIds && (props.onMouseMove || props.onMouseEnter || props.onMouseLeave);
        if (shouldTrackHoveredFeatures) {
            var eventType = e.type;
            var wasHovering = ((_a = this._hoveredFeatures) === null || _a === void 0 ? void 0 : _a.length) > 0;
            var features = void 0;
            if (eventType === 'mousemove') {
                try {
                    features = this._map.queryRenderedFeatures(e.point, {
                        layers: props.interactiveLayerIds
                    });
                }
                catch (_b) {
                    features = [];
                }
            }
            else {
                features = [];
            }
            var isHovering = features.length > 0;
            if (!isHovering && wasHovering) {
                e.type = 'mouseleave';
                this._onPointerEvent(e);
            }
            this._hoveredFeatures = features;
            if (isHovering && !wasHovering) {
                e.type = 'mouseenter';
                this._onPointerEvent(e);
            }
            e.type = eventType;
        }
        else {
            this._hoveredFeatures = null;
        }
    };
    Mapbox.prototype._fireEvent = function (baseFire, event, properties) {
        var map = this._map;
        var tr = map.transform;
        var eventType = typeof event === 'string' ? event : event.type;
        switch (eventType) {
            case 'resize':
                this._renderTransform.resize(tr.width, tr.height);
                break;
            case 'move':
                this._updateViewState(this.props, false);
                break;
            case 'mousemove':
            case 'mouseout':
                // @ts-ignore
                this._updateHover(event);
                break;
            default:
        }
        if (eventType in cameraEvents) {
            if (typeof event === 'object') {
                event.viewState = (0, transform_1.transformToViewState)(tr);
            }
            if (this._map.isMoving()) {
                // Replace map.transform with ours during the callbacks
                map.transform = this._renderTransform;
                baseFire.call(map, event, properties);
                map.transform = tr;
                return map;
            }
        }
        baseFire.call(map, event, properties);
        return map;
    };
    Mapbox.prototype._render = function (baseRender, arg) {
        var map = this._map;
        this._inRender = true;
        if (this._moved) {
            this._internalUpdate = true;
            map.fire('move');
            if (this._zoomed) {
                map.fire('zoom');
                this._zoomed = false;
            }
            if (this._rotated) {
                map.fire('rotate');
                this._rotated = false;
            }
            if (this._pitched) {
                map.fire('pitch');
                this._pitched = false;
            }
            this._moved = false;
            this._internalUpdate = false;
        }
        // map.transform will be swapped out in _onBeforeRender
        var tr = map.transform;
        baseRender.call(map, arg);
        map.transform = tr;
        this._inRender = false;
        // We do not allow props to change during a render
        // When render is done, apply any pending changes
        if (this._nextProps) {
            this.setProps(this._nextProps);
            this._nextProps = null;
        }
    };
    Mapbox.prototype._onBeforeRepaint = function () {
        // Make sure camera matches the current props
        this._map.transform = this._renderTransform;
    };
    return Mapbox;
}());
exports.default = Mapbox;
/**
 * Access token can be provided via one of:
 *   mapboxAccessToken prop
 *   access_token query parameter
 *   MapboxAccessToken environment variable
 *   REACT_APP_MAPBOX_ACCESS_TOKEN environment variable
 * @returns access token
 */
function getAccessTokenFromEnv() {
    var accessToken = null;
    /* global location, process */
    if (typeof location !== 'undefined') {
        var match = /access_token=([^&\/]*)/.exec(location.search);
        accessToken = match && match[1];
    }
    // Note: This depends on bundler plugins (e.g. webpack) importing environment correctly
    try {
        accessToken = accessToken || process.env.MapboxAccessToken;
    }
    catch (_a) {
        // ignore
    }
    try {
        accessToken = accessToken || process.env.REACT_APP_MAPBOX_ACCESS_TOKEN;
    }
    catch (_b) {
        // ignore
    }
    return accessToken;
}
//# sourceMappingURL=mapbox.js.map